#!/usr/bin/node

var fs = require("fs");

var input = fs.readFileSync("vstup.txt", { encoding: "utf8" }).trim().split("\n");

function Tile(x, y, type) {
	this.type = type;
	this.x = x;
	this.y = y;
}

Tile.prototype.swap = function(t2) {
	board[t2.y][t2.x] = this;
	board[this.y][this.x] = t2;

	var oldx = this.x, oldy = this.y;

	this.y = t2.y;
	this.x = t2.x;
	t2.y = oldy;
	t2.x = oldx;
};

var stepsLeft = input.shift().trim();
var board = input.map(function(row, i) {
	return row.trim().split("").map(function(c, j) {
		return new Tile(j, i, parseInt(c, 10) - 1);
	});
});

function findPossibleMoves() {
	var moves = [];

	for(var i = 0; i < 8; i++) {
		for(var j = 0; j < 8; j++) {
			//console.log(board[i]);

			var t = board[i][j].type;

			// row

			// xx?
			try {
				if(board[i][j + 1] && t === board[i][j + 1].type) {
					if(board[i - 1] && board[i - 1][j + 2] && t === board[i - 1][j + 2].type) {
						moves.push([ board[i][j + 2], board[i - 1][j + 2] ]);
					}
					if(board[i + 1] && board[i + 1][j + 2] && t === board[i + 1][j + 2].type) {
						moves.push([ board[i][j + 2], board[i + 1][j + 2] ]);
					}
					if(board[i] && board[i][j + 3] && t === board[i][j + 3].type) {
						moves.push([ board[i][j + 2], board[i][j + 3] ]);
					}
				}
			} catch(err) {}

			// x?x
			try {
				if(board[i][j + 2] && board[i][j].type === board[i][j + 2].type) {
					if(board[i - 1] && board[i - 1][j + 1] && t === board[i - 1][j + 1].type) {
						moves.push([ board[i][j + 1], board[i - 1][j + 1] ]);
					}
					if(board[i + 1] && board[i + 1][j + 1] && t === board[i + 1][j + 1].type) {
						moves.push([ board[i][j + 1], board[i + 1][j + 1] ]);
					}
				}
			} catch(err) {}

			// ?xx
			try {
				if(board[i][j + 1] && board[i][j + 2] && board[i][j + 1].type === board[i][j + 2].type) {
					if(board[i][j - 1] && board[i][j + 1].type === board[i][j - 1].type) {
						moves.push([ board[i][j], board[i][j - 1] ]);
					}
					if(board[i - 1] && board[i][j + 1].type === board[i - 1][j].type) {
						moves.push([ board[i][j], board[i - 1][j] ]);
					}
					if(board[i + 1] && board[i][j + 1].type === board[i + 1][j].type) {
						moves.push([ board[i][j], board[i + 1][j] ]);
					}
				}
			} catch(err) {}

			// col

			// xx?
			try {
				if(board[i + 1] && t === board[i + 1][j].type) {
					if(board[i + 2] && board[i + 2][j - 1] && t === board[i + 2][j - 1].type) {
						moves.push([ board[i + 2][j], board[i + 2][j - 1] ]);
					}
					if(board[i + 2] && board[i + 2][j + 1] && t === board[i + 2][j + 1].type) {
						moves.push([ board[i + 2][j], board[i + 2][j + 1] ]);
					}
					if(board[i + 3] && t === board[i + 3][j].type) {
						moves.push([ board[i + 2][j], board[i + 3][j] ]);
					}
				}
			} catch(err) {}

			// x?x
			try {
				if(board[i + 2] && board[i][j].type === board[i + 2][j].type) {
					if(board[i + 1] && board[i + 1][j - 1] && t === board[i + 1][j - 1].type) {
						moves.push([ board[i + 1][j], board[i + 1][j - 1] ]);
					}
					if(board[i + 1] && board[i + 1][j + 1] && t === board[i + 1][j + 1].type) {
						moves.push([ board[i + 1][j], board[i + 1][j + 1] ]);
					}
				}
			} catch(err) {}

			// ?xx
			try {
				if(board[i + 1] && board[i + 2] && board[i + 1][j].type === board[i + 2][j].type) {
					if(board[i - 1] && board[i + 1][j].type === board[i - 1][j].type) {
						moves.push([ board[i][j], board[i - 1][j] ]);
					}
					if(board[i][j - 1] && board[i + 1][j].type === board[i][j - 1].type) {
						moves.push([ board[i][j], board[i][j - 1] ]);
					}
					if(board[i][j + 1] && board[i + 1][j].type === board[i][j + 1].type) {
						moves.push([ board[i][j], board[i][j + 1] ]);
					}
				}	
			} catch(err) {}
		}
	}

	return moves;
}

function findMatches() {
	var matches = [];
	var tileIndex = {};

	function index(tile, match) {
		if(tileIndex[tile.x + ":" + tile.y])
			tileIndex[tile.x + ":" + tile.y].push(match);
		else
			tileIndex[tile.x + ":" + tile.y] = [match];
	}

	var i, j, match;

	// find x matches
	for(i = 0; i < 8; i++) {
		for(j = 0; j < 6; j++) {

			if(board[i][j].type === board[i][j + 1].type && board[i][j + 1].type === board[i][j + 2].type) {
				matches.push(match = {
					type: "row",
					x: j,
					y: i,
					tiles: [ board[i][j], board[i][j + 1], board[i][j + 2] ]
				});

				index(board[i][j], match);
				index(board[i][j + 1], match);
				index(board[i][j + 2], match);
			}
		
		}
	}

	// find y matches
	for(i = 0; i < 6; i++) {
		for(j = 0; j < 8; j++) {

			if(board[i][j].type === board[i + 1][j].type && board[i + 1][j].type === board[i + 2][j].type) {
				matches.push(match = {
					type: "col",
					x: j,
					y: i,
					tiles: [ board[i][j], board[i + 1][j], board[i + 2][j] ]
				});

				index(board[i][j], match);
				index(board[i + 1][j], match);
				index(board[i + 2][j], match);
			}
		
		}
	}

	return [ matches, tileIndex ];
}

function cloneBoard() {
	var newBoard = [];
	for(var i = 0; i < 8; i++) {
		newBoard[i] = [];
		for(var j = 0; j < 8; j++) {
			newBoard[i][j] = new Tile(j, i, board[i][j].type);
		}
	}
	return newBoard;
}

function simulateMove(move) {
	var oldboard = JSON.stringify(board);

	move[0].swap(move[1]);
	var score = findMatches()[0].length;
	move[1].swap(move[0]);


	if(oldboard !== JSON.stringify(board))
		console.error("board dont match");

	return score;
}

var moves = findPossibleMoves();

var maxScore = 0, maxMove = null;
for(var i = 0; i < moves.length; i++) {
	var score = simulateMove(moves[i]);

	if(score > maxScore) {
		maxScore = score;
		maxMove = moves[i];
	}
}

var data = [
	maxMove[0].x + 1,
	maxMove[0].y + 1,
	maxMove[1].x + 1,
	maxMove[1].y + 1
];

fs.writeFileSync("vystup.txt", data.join(" "), { encoding: "utf8" });
