var $ = document.querySelector.bind(document);

$("#settings").onsubmit = function(e) {
	e.preventDefault();

	var board = $("#board"), timeLimit = $("#timeLimit");

	if(timeLimit.value !== "") {
		var m = timeLimit.value.match(/^([0-9]+):([0-9]+)$/);
		if(!m)
			return alert("Časový limit musí být ve formátu mm:ss");
		var min = parseInt(m[1], 10), sec = parseInt(m[2], 10);

		if(isNaN(min) || isNaN(sec) || sec > 59)
			return alert("Neplatný časový limit.");

		timeLimit = min * 60 + sec;
	} else timeLimit = Infinity;		

	if(board.files && board.files.length) {
		var reader = new FileReader();
		reader.onload = function() {
			startGame({
				board: reader.result,
				timeLimit: timeLimit
			});
		};
		reader.onerror = function() {
			alert("Chyba při čtení souboru.");
		};
		reader.readAsText(board.files[0]);
	} else {
		startGame({
			board: null,
			timeLimit: timeLimit
		});
	}
};

$("#hint").onclick = function(e) {
	e.preventDefault();

	var moves = findPossibleMoves();
	if(!moves.length)
		return alert("Nejsou možné žádné další tahy.");

	for(var i = 0; i < moves.length; i++) {
		moves[i][0].hilit = moves[i][1].hilit = true;
	}

	setTimeout(function() {
		for(var i = 0; i < moves.length; i++) {
			moves[i][0].hilit = moves[i][1].hilit = false;
		}
	}, 2500);
};

var BOARD_WIDTH = 8, BOARD_HEIGHT = 8, TILE_SIZE = 48;
var COLORS = [ "#FF0000", "#00FF00", "#0000FF", "#FFFFFF", "#FFFF00", "#FF00FF" ];

var board, activeTile, spareTiles, spareTileIndexes;

function Tile(x, y, type, active) {
	this.type = type;
	this.active = !!active;
	this.x = x;
	this.y = y;
	this.valid = true;
	this.hilit = false;

	this.animation = null;
}

Tile.makeRandom = function(x, y) {
	return new Tile(x, y, Math.floor(Math.random() * COLORS.length), false);
};

Tile.prototype.getColor = function() {
	return COLORS[this.type];
};

Tile.prototype.setActive = function(t) {
	if(t === undefined)
		t = true;

	this.active = t;
	if(t)
		activeTile = this;
	else if(activeTile === this)
		activeTile = null;
};

Tile.prototype.swap = function(t2) {
	board[t2.y][t2.x] = this;
	board[this.y][this.x] = t2;

	var oldx = this.x, oldy = this.y;

	this.y = t2.y;
	this.x = t2.x;
	t2.y = oldy;
	t2.x = oldx;

	this.hilit = false;
	t2.hilit = false;
};

Tile.prototype.remove = function() {
	if(!this.valid)
		return;
	
	this.valid = false;
};

Tile.prototype.render = function(time) {
	if(!this.valid)
		return;

	var x = this.x * TILE_SIZE,
		y = this.y * TILE_SIZE;

	if(this.animation) {
		var progress = Math.min(Math.max((time - this.animation.start) / this.animation.duration, 0), 1);
		y = (this.animation.startY + (this.animation.endY - this.animation.startY) * progress) * TILE_SIZE;

		if(progress === 1)
			this.animation = null;
	}

	ctx.fillStyle = this.getColor();
	ctx.beginPath();
	ctx.arc(x + TILE_SIZE / 2, y + TILE_SIZE / 2, TILE_SIZE / 2, 0, Math.PI * 2);
	ctx.fill();
	ctx.closePath();
};

Tile.prototype.renderOutline = function(time) {
	if(!this.valid)
		return;

	var x = this.x * TILE_SIZE,
		y = this.y * TILE_SIZE;

	if(this.active) {
		ctx.strokeStyle = "#FFFFFF";
		ctx.lineWidth = 3;
		ctx.strokeRect(x, y, TILE_SIZE, TILE_SIZE);
	}

	if(this.hilit) {
		ctx.strokeStyle = "#FFFF00";
		ctx.lineWidth = 3;
		ctx.strokeRect(x, y, TILE_SIZE, TILE_SIZE);
	}
};

Tile.prototype.animate = function(targetY, startY) {
	var startY2 = startY !== undefined ? startY : this.y;

	if(this.animation) {
		this.animation.endY = targetY;
		if(startY !== undefined)
			this.animation.startY = startY;

		this.animation.duration = Math.abs(this.animation.startY - targetY) * 100;
	} else {
		this.animation = {
			start: Date.now(),
			duration: Math.abs(startY2 - targetY) * 100,
			startY: startY2,
			endY: targetY
		};
	}
};

function createBoard() {
	var board = [];
	for(var i = 0; i < BOARD_HEIGHT; i++) {
		board[i] = [];
		for(var j = 0; j < BOARD_WIDTH; j++) {
			board[i][j] = Tile.makeRandom(j, i);
		}
	}
	return board;
}

var running = false, points = 0, pointsTick = 0, timeEnd = Infinity;

function startGame(options) {
	$("#settings").reset();
	initCanvas();

	board = createBoard();

	if(options.board) {
		var boardData = options.board.trim().split("\r\n");

		console.log(boardData);

		if(boardData.length !== BOARD_WIDTH + BOARD_HEIGHT) {
			alert("Neplatná herní plocha!");
			return;
		}

		var i;
		for(i = 0; i < BOARD_HEIGHT; i++) {
			for(var j = 0; j < BOARD_WIDTH; j++) {
				board[i][j].type = parseInt(boardData[i].charAt(j), 10) - 1;
			}
		}

		spareTiles = [];
		for(i = 8; i < 16; i++) {
			spareTiles[i - 8] = boardData[i].trim().split("").map(function(r) {
				return parseInt(r, 10) - 1;
			});
		}

	} else {
		spareTiles = [];
		for(var i = 0; i < BOARD_WIDTH; i++) {
			spareTiles[i] = [];
			for(var j = 0; j < 1024; j++)
				spareTiles[i].push(Math.floor(Math.random() * COLORS.length));
		}
	}

	spareTileIndexes = [];
	for(var i = 0; i < BOARD_WIDTH; i++) {
		spareTileIndexes[i] = 0;
	}

	doGameTick();
	points = pointsTick = 0;

	timeEnd = Number.isFinite(options.timeLimit) ? Date.now() + options.timeLimit * 1000 : Infinity;
	updateStatus();

	running = true;
	requestAnimationFrame(function render() {
		renderBoard(board);
		renderFill();
		
		requestAnimationFrame(render);
	});
	setInterval(updateStatus, 500);
}

function findPossibleMoves() {
	var moves = [];

	// find in row
	for(var i = 0; i < 8; i++) {
		for(var j = 0; j < 8; j++) {
			//console.log(board[i]);

			var t = board[i][j].type;

			// xx?
			try {
				if(board[i][j + 1] && t === board[i][j + 1].type) {
					if(board[i - 1] && board[i - 1][j + 2] && t === board[i - 1][j + 2].type) {
						moves.push([ board[i][j + 2], board[i - 1][j + 2] ]);
					}
					if(board[i + 1] && board[i + 1][j + 2] && t === board[i + 1][j + 2].type) {
						moves.push([ board[i][j + 2], board[i + 1][j + 2] ]);
					}
					if(board[i] && board[i][j + 3] && t === board[i][j + 3].type) {
						moves.push([ board[i][j + 2], board[i][j + 3] ]);
					}
				}
			} catch(err) {}

			// x?x
			try {
				if(board[i][j + 2] && board[i][j].type === board[i][j + 2].type) {
					if(board[i - 1] && board[i - 1][j + 1] && t === board[i - 1][j + 1].type) {
						moves.push([ board[i][j + 1], board[i - 1][j + 1] ]);
					}
					if(board[i + 1] && board[i + 1][j + 1] && t === board[i + 1][j + 1].type) {
						moves.push([ board[i][j + 1], board[i + 1][j + 1] ]);
					}
				}
			} catch(err) {}

			// ?xx
			try {
				if(board[i][j + 1] && board[i][j + 2] && board[i][j + 1].type === board[i][j + 2].type) {
					if(board[i][j - 1] && board[i][j + 1].type === board[i][j - 1].type) {
						moves.push([ board[i][j], board[i][j - 1] ]);
					}
					if(board[i - 1] && board[i][j + 1].type === board[i - 1][j].type) {
						moves.push([ board[i][j], board[i - 1][j] ]);
					}
					if(board[i + 1] && board[i][j + 1].type === board[i + 1][j].type) {
						moves.push([ board[i][j], board[i + 1][j] ]);
					}
				}
			} catch(err) {}

			// xx?
			try {
				if(board[i + 1] && t === board[i + 1][j].type) {
					if(board[i + 2] && board[i + 2][j - 1] && t === board[i + 2][j - 1].type) {
						moves.push([ board[i + 2][j], board[i + 2][j - 1] ]);
					}
					if(board[i + 2] && board[i + 2][j + 1] && t === board[i + 2][j + 1].type) {
						moves.push([ board[i + 2][j], board[i + 2][j + 1] ]);
					}
					if(board[i + 3] && t === board[i + 3][j].type) {
						moves.push([ board[i + 2][j], board[i + 3][j] ]);
					}
				}
			} catch(err) {}

			// x?x
			try {
				if(board[i + 2] && board[i][j].type === board[i + 2][j].type) {
					if(board[i + 1] && board[i + 1][j - 1] && t === board[i + 1][j - 1].type) {
						moves.push([ board[i + 1][j], board[i + 1][j - 1] ]);
					}
					if(board[i + 1] && board[i + 1][j + 1] && t === board[i + 1][j + 1].type) {
						moves.push([ board[i + 1][j], board[i + 1][j + 1] ]);
					}
				}
			} catch(err) {}

			// ?xx
			try {
				if(board[i + 1] && board[i + 2] && board[i + 1][j].type === board[i + 2][j].type) {
					if(board[i - 1] && board[i + 1][j].type === board[i - 1][j].type) {
						moves.push([ board[i][j], board[i - 1][j] ]);
					}
					if( board[i][j - 1] && board[i + 1][j].type === board[i][j - 1].type) {
						moves.push([ board[i][j], board[i][j - 1] ]);
					}
					if(board[i][j + 1] && board[i + 1][j].type === board[i][j + 1].type) {
						moves.push([ board[i][j], board[i][j + 1] ]);
					}
				}	
			} catch(err) {}
		}
	}

	return moves;
}

function findMatches() {
	var matches = [];
	var tileIndex = {};

	function index(tile, match) {
		if(tileIndex[tile.x + ":" + tile.y])
			tileIndex[tile.x + ":" + tile.y].push(match);
		else
			tileIndex[tile.x + ":" + tile.y] = [match];
	}

	var i, j, match;

	// find x matches
	for(i = 0; i < BOARD_HEIGHT; i++) {
		for(j = 0; j < BOARD_WIDTH - 2; j++) {

			if(board[i][j].type === board[i][j + 1].type && board[i][j + 1].type === board[i][j + 2].type) {
				matches.push(match = {
					type: "row",
					x: j,
					y: i,
					tiles: [ board[i][j], board[i][j + 1], board[i][j + 2] ]
				});

				index(board[i][j], match);
				index(board[i][j + 1], match);
				index(board[i][j + 2], match);
			}
		
		}
	}

	// find y matches
	for(i = 0; i < BOARD_HEIGHT - 2; i++) {
		for(j = 0; j < BOARD_WIDTH; j++) {

			if(board[i][j].type === board[i + 1][j].type && board[i + 1][j].type === board[i + 2][j].type) {
				matches.push(match = {
					type: "col",
					x: j,
					y: i,
					tiles: [ board[i][j], board[i + 1][j], board[i + 2][j] ]
				});

				index(board[i][j], match);
				index(board[i + 1][j], match);
				index(board[i + 2][j], match);
			}
		
		}
	}

	return [ matches, tileIndex ];
}

function handleGameEnd() {
	running = false;
	alert("Konec hry! Bodů celkem: " + points);
}

function doGameTick(matches) {
	console.log("doGameTick");

	if(!matches)
		matches = findMatches();

	matches = matches[0];

	console.log(matches);

	if(!matches.length)
		return false;

	// calculate points
	pointsTick += matches.length * 3;

	var i, j;

	// remove triplets
	for(i = 0; i < matches.length; i++) {
		for(j = 0; j < 3; j++)
			matches[i].tiles[j].remove();
	}

	// do gravity
	console.log("---");
	var changed;
	var a = 0;
	do {
		changed = 0;
		for(i = 1; i < BOARD_HEIGHT; i++) {
			for(j = 0; j < BOARD_WIDTH; j++) {
				if(!board[i][j].valid && board[i - 1][j].valid) {
					changed++;

					board[i - 1][j].animate(i);
					board[i][j].swap(board[i - 1][j]);
				}

			}
		}
		console.log(changed, a);
	} while(changed && a++ < 30);

	for(i = 1; i < BOARD_HEIGHT; i++) {
		for(j = 0; j < BOARD_WIDTH; j++) {
			if(!board[i][j].valid && board[i - 1][j].valid)
				console.error("LOOP!!!");
		}
	}

	console.log("...");

	//replace invalid tiles with new ones
	for(i = BOARD_HEIGHT; i--;) {
		for(j = 0; j < BOARD_WIDTH; j++) {
			if(!board[i][j].valid) {
				board[i][j].valid = true;
				board[i][j].type = spareTiles[j][spareTileIndexes[j]];
				board[i][j].animate(i, -1);

				spareTileIndexes[j] ++;
				if(spareTileIndexes[j] >= spareTiles[j].length)
					spareTileIndexes[j] = 0;
			}
		}
	}

	return doGameTick();
}

function swapTiles(t1, t2) {
	console.log("swapTiles");

	pointsTick = 0;

	t1.swap(t2);

	var matches = findMatches();

	if(!matches[0].length)
		return t2.swap(t1);

	doGameTick(matches);
	points += pointsTick;

	if(!findPossibleMoves().length)
		handleGameEnd();
}

function isNear(t1, t2) {
	return (Math.abs(t1.x - t2.x) === 1 && Math.abs(t1.y - t2.y) === 0) || (Math.abs(t1.x - t2.x) === 0 && Math.abs(t1.y - t2.y) === 1);
}

function handleTileClick(tile) {
	if(tile.active)
		return;

	if(activeTile) {
		var oldactive = activeTile;
		oldactive.setActive(false);

		if(isNear(tile, oldactive)) {
			swapTiles(tile, oldactive);
		} else {
			tile.setActive();
		}
	} else {
		tile.setActive();
	}
}

// rendering
var canvas = $("#game"),
	ctx = canvas.getContext("2d");

function initCanvas() {
	canvas.width = BOARD_WIDTH * TILE_SIZE;
	canvas.height = BOARD_HEIGHT * TILE_SIZE;
}

function renderBoard(board) {
	var time = Date.now();

	ctx.fillStyle = "#000000";
	ctx.fillRect(0, 0, canvas.width, canvas.height);

	var i, j;
	for(i = 0; i < board.length; i++) {
		for(j = 0; j < board.length; j++) {
			board[i][j].render(time);
		}
	}

	for(i = 0; i < board.length; i++) {
		for(j = 0; j < board.length; j++) {
			board[i][j].renderOutline(time);
		}
	}
}

canvas.onclick = function(e) {
	e.preventDefault();

	if(!running)
		return;

	var tileX = Math.floor(e.offsetX / TILE_SIZE),
		tileY = Math.floor(e.offsetY / TILE_SIZE);

	handleTileClick(board[tileY][tileX]);
};

function updateStatus() {
	if(!running)
		return;

	var now = Date.now();

	if(now >= timeEnd) {
		running = false;
		handleGameEnd();
	}

	$("#pointsTick").textContent = pointsTick;
	$("#pointsGame").textContent = points;

	if(!Number.isFinite(timeEnd))
		$("#time").textContent = "∞";
	else if(timeEnd <= now)
		$("#time").textContent = "00:00";
	else {
		var sec = Math.round((timeEnd - now) / 1000),
			min = Math.floor(sec / 60);
		sec = sec % 60;

		sec = sec < 10 ? "0" + sec.toString() : sec.toString();
		min = min < 10 ? "0" + min.toString() : min.toString();

		$("#time").textContent = min + ":" + sec;		
	}
}

var fillCanvas = $("#fill"),
	fillCtx = fillCanvas.getContext("2d");
function renderFill() {
	fillCanvas.width = BOARD_WIDTH * TILE_SIZE;
	fillCanvas.height = TILE_SIZE;

	fillCtx.fillStyle = "#000000";
	fillCtx.fillRect(0, 0, fillCanvas.width, fillCanvas.height);

	for(var i = 0; i < BOARD_WIDTH; i++) {
		var s = spareTiles[i][spareTileIndexes[i]];

		var x = i * TILE_SIZE;

		fillCtx.fillStyle = COLORS[s];
		fillCtx.beginPath();
		fillCtx.arc(x + TILE_SIZE / 2, TILE_SIZE / 2, TILE_SIZE / 2, 0, Math.PI * 2);
		fillCtx.fill();
		fillCtx.closePath();
	}
}